@@ -4,6 +4,7 @@
 
 from dataclasses import dataclass
 import os, types
+import math
 import torch
 import torch.nn as nn
 import torch.nn.functional as F
@@ -308,13 +309,58 @@ class CausalSparseAttention(nn.Module):
         # regularization
         self.n_head = config.n_head
         self.n_embd = config.n_embd
-        self.moba_chunk_size = config.moba_chunk_size
-        self.moba_topk = config.moba_topk
-        self.window_size = config.moba_chunk_size * config.moba_topk
+        self.attn_chunk_size = config.attn_chunk_size
+        self.attn_topk = config.attn_topk
+        self.short_sequence_criteria = config.short_sequence_criteria
+        # kv cache management
+        self.max_kv_cache_size = config.max_kv_cache_size # condition that trigger the cache management
+        self.kv_cache_window_size = config.kv_cache_window_size # observation window size
+        self.min_kv_cache_size = config.min_kv_cache_size # minimum kv cache size
+        self.prefill_attn_mode = config.prefill_attn_mode # attention mode for prefill
+        self.decoding_attn_mode = config.decoding_attn_mode # attention mode for decoding
+
+    def update_kv_cache(self, q, k, v):
+        '''
+        input: q, k, v: (B, QT, C), (B, KT, C)
+        output: k_past_compress, v_past_compress: (B, min_kv_cache_size, C)
+        '''
+        B, QT, C = q.size()
+        KT, VT = k.size(1), v.size(1)
+        assert KT == VT, "key and value must have the same length"
+        # check if the cache size is larger than the minimum size
+        if self.min_kv_cache_size >= KT:
+            return k, v
+        # split k, v into past and current
+        past_KT = KT - self.kv_cache_window_size
+        k_past, k_cur = k[:, :past_KT, :], k[:, past_KT: :]
+        v_past, v_cur = v[:, :past_KT, :], v[:, past_KT:, :]
+        q = q.view(B, QT, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, QT, hs)
+        k_past = k_past.view(B, past_KT, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, past_KT, hs)
+        v_past = v_past.view(B, past_KT, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, past_KT, hs)
+        # calculate attention weights
+        B, NH, QT, HS = q.size()
+        attn_weights = torch.matmul(q, k_past.transpose(2, 3)) / math.sqrt(HS) # (B, NH, QT, past_KT)
+        attn_weights = F.softmax(attn_weights, dim=-1) # (B, NH, QT, past_KT)
+        vote = attn_weights.sum(dim=-2) # (B, NH, past_KT) Sum the weight along the query dimension
+        assert self.min_kv_cache_size - self.kv_cache_window_size > 0
+        keep_KT = min(self.min_kv_cache_size - self.kv_cache_window_size, past_KT)
+        indices = vote.topk(keep_KT, dim=-1).indices
+        # Expand the indices to match the head dimension for gathering
+        indices = indices.unsqueeze(-1).expand(-1, -1, -1, HS)
+        k_past_compress = torch.gather(k_past, 2, indices)
+        v_past_compress = torch.gather(v_past, 2, indices)
+        # Reshape back to (B, KT, C)
+        k_past_compress = k_past_compress.transpose(1, 2).contiguous().view(B, -1, C) # (B, keep_KT, C)
+        v_past_compress = v_past_compress.transpose(1, 2).contiguous().view(B, -1, C) # (B, keep_KT, C)
+        # 
+        k_cache = torch.cat([k_past_compress, k_cur], dim=1) # (B, keep_KT, C)
+        v_cache = torch.cat([v_past_compress, v_cur], dim=1)
+        return k_cache, v_cache
 
     def forward_one(self, x, k_cache, v_cache):
         ''' used for decode, only one token at a time
-        input: x: (C)
+        input: x: (C) and k, v cache (1, CT, C)
+        output: y: (C) and k, v cache (1, CT+1, C)
         '''
         if len(x.shape) == 1:
             x = x.unsqueeze(0).unsqueeze(0) # (C) -> (1, 1, C)
@@ -322,31 +368,39 @@ class CausalSparseAttention(nn.Module):
         assert x.size(1) == 1, "sequence length must be 1"
         C = x.size(-1)
         q, k, v = self.receptance(x), self.key(x), self.value(x)
-        # split kv cache into two parts
-        KT = k_cache.size(1)
-        if KT <= self.window_size: # for short sequence, use full attention
-            k_cache = torch.cat((k_cache, k), dim=1)
-            v_cache = torch.cat((v_cache, v), dim=1)
-            y = F.scaled_dot_product_attention(q, k_cache, v_cache)
-            y = y.transpose(1, 2).contiguous().view(C) # (1, 1, C) -> (C)
+        # manage k, v cache, only when not use full attention
+        if self.deocoding_attn_mode != 'full' and k_cache.size(1) > self.max_kv_cache_size:
+            k_cache, v_cache = self.update_kv_cache(q, k_cache, v_cache)
+        CT = k_cache.size(1)
+        # apply the attention, only decoding consider the full attention mode
+        if CT <= self.short_sequence_criteria or self.deocoding_attn_mode == 'full': # for short sequence, use full attention
+            k_cache = torch.cat((k_cache, k), dim=1) # update k cache
+            v_cache = torch.cat((v_cache, v), dim=1) # update v cache
+            q = q.view(1, 1, self.n_head, C // self.n_head).transpose(1, 2) # (1, 1, C) -> (1, nh, 1, hs)
+            k_comb = k_cache.view(1, CT+1, self.n_head, C // self.n_head).transpose(1, 2) # (1, nh, CT+1, hs)
+            v_comb = v_cache.view(1, CT+1, self.n_head, C // self.n_head).transpose(1, 2) # (1, nh, CT+1, hs)
+            # !!! super be careful here, the attention is not causal !!!
+            # is_causal=True -> torch.ones(L, S, dtype=torch.bool).tril(diagonal=0) -> [1， S] -> [1, 0, 0, 0, ...]
+            y = F.scaled_dot_product_attention(q, k_comb, v_comb)
+            y = y.transpose(1, 2).contiguous().view(C) # (1, nh, 1, hs) -> (1, 1, nh, hs) -> (C)
             y = self.output(y)
             return y, k_cache, v_cache
         else:
-            reminder = k_cache.size(1) % self.moba_chunk_size
-            k_chunk, k_reminder = k_cache[:, :KT-reminder, :], k_cache[:, KT-reminder:, :]
-            v_chunk, v_reminder = v_cache[:, :KT-reminder, :], v_cache[:, KT-reminder:, :]
+            reminder = CT % self.attn_chunk_size
+            k_chunk, k_reminder = k_cache[:, :CT-reminder, :], k_cache[:, CT-reminder:, :]
+            v_chunk, v_reminder = v_cache[:, :CT-reminder, :], v_cache[:, CT-reminder:, :]
             # split k, v into chunks
-            num_chunks = k_chunk.size(1) // self.moba_chunk_size
-            k_chunk = k_chunk.view(1, num_chunks, self.moba_chunk_size, C)
-            v_chunk = v_chunk.view(1, num_chunks, self.moba_chunk_size, C)
+            num_chunks = k_chunk.size(1) // self.attn_chunk_size
+            k_chunk = k_chunk.view(1, num_chunks, self.attn_chunk_size, C)
+            v_chunk = v_chunk.view(1, num_chunks, self.attn_chunk_size, C)
             # get chunk key
             chunk_key = k_chunk.mean(dim=2) # (1, num_chunks, C)
             # get top-k chunk
             chunk_score = torch.einsum('bqc,bkc->bqk', q, chunk_key) # (1, 1, num_chunks)
-            topk_chunk_indices = torch.topk(chunk_score, self.moba_topk, dim=-1).indices.squeeze() # (moba_topk)
+            topk_chunk_indices = torch.topk(chunk_score, self.attn_topk, dim=-1).indices.squeeze() # (attn_topk)
             # get top-k chunk key and value
-            topk_k = k_chunk[:, topk_chunk_indices].view(1, -1, C) # (1, moba_topk * moba_chunk_size, C)
-            topk_v = v_chunk[:, topk_chunk_indices].view(1, -1, C) # (1, moba_topk * moba_chunk_size, C)
+            topk_k = k_chunk[:, topk_chunk_indices].view(1, -1, C) # (1, attn_topk * attn_chunk_size, C)
+            topk_v = v_chunk[:, topk_chunk_indices].view(1, -1, C) # (1, attn_topk * attn_chunk_size, C)
             # combine with reminder and current k, v
             k_comb = torch.cat((topk_k, k_reminder, k), dim=1)
             v_comb = torch.cat((topk_v, v_reminder, v), dim=1)
@@ -355,10 +409,10 @@ class CausalSparseAttention(nn.Module):
             q = q.view(1, 1, self.n_head, C // self.n_head).transpose(1, 2)
             k_comb = k_comb.view(1, new_T, self.n_head, C // self.n_head).transpose(1, 2)
             v_comb = v_comb.view(1, new_T, self.n_head, C // self.n_head).transpose(1, 2)
+            # !!! super be careful, the attention here can not be causal !!!
             y = F.scaled_dot_product_attention(q, k_comb, v_comb)
             y = y.transpose(1, 2).contiguous().view(C) # (1, 1, C) -> (C)
             y = self.output(y)
-            # update k, v cache
             k_cache = torch.cat((k_cache, k), dim=1)
             v_cache = torch.cat((v_cache, v), dim=1)
             return y, k_cache, v_cache
@@ -371,46 +425,54 @@ class CausalSparseAttention(nn.Module):
         if len(x.shape) == 2:
             x = x.unsqueeze(0) # (T, C) -> (1, T, C)
         B, T, C = x.size()
+        # manage k, v cache, only when not use full attention
+        if self.attn_mode != 'full' and k_cache.size(1) > self.max_kv_cache_size:
+            k_cache, v_cache = self.update_kv_cache(x, k_cache, v_cache)
         CT = k_cache.size(1) # cache seq length
-        if (T+CT) <= self.window_size: # for short sequence, use full attention
+        # apply the attention
+        if (T+CT) <= self.short_sequence_criteria or self.prefill_attn_mode == 'full': # for short sequence, use full attention
             q, k, v = self.receptance(x), self.key(x), self.value(x)
+            # prefix mask
+            causal_mask = torch.ones((T, T), dtype=torch.bool, device=x.device).tril(diagonal=0)
+            cache_mask = torch.ones((T, CT), dtype=torch.bool, device=x.device)
+            prefix_mask = torch.cat((cache_mask, causal_mask), dim=1)
             k_cache = torch.cat((k_cache, k), dim=1) # update k cache
             v_cache = torch.cat((v_cache, v), dim=1) # update v cache
             q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
-            k = k_cache.view(B, CT+T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, CT+T, hs)
-            v = v_cache.view(B, CT+T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, CT+T, hs)
-            y = F.scaled_dot_product_attention(q, k, v, is_causal=True) # flash attention
+            k_comb = k_cache.view(B, CT+T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, CT+T, hs)
+            v_comb = v_cache.view(B, CT+T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, CT+T, hs)
+            y = F.scaled_dot_product_attention(q, k_comb, v_comb, attn_mask=prefix_mask)
             y = y.transpose(1, 2).contiguous().view(B, T, C) # re-assemble all head outputs side by side
             # output projection
             y = self.output(y).squeeze(0) # (1, T, C) -> (T, C)
             return y, k_cache, v_cache
         else: # chunked attention ignoring the kv cache
             # pad on right side to match the chunk size
-            pad_size = self.moba_chunk_size - (T % self.moba_chunk_size)
-            if pad_size != self.moba_chunk_size:
+            pad_size = self.attn_chunk_size - (T % self.attn_chunk_size)
+            if pad_size != self.attn_chunk_size:
                 zero_pad = torch.zeros((B, pad_size, C), dtype=x.dtype, device=x.device)
                 x = torch.cat((x, zero_pad), dim=1)
                 T = x.size(1) # update T after padding
             # fold x into chunks, merge with batch size dim
-            num_chunks = T // self.moba_chunk_size
-            x = x.view(B * num_chunks, self.moba_chunk_size, C)
+            num_chunks = T // self.attn_chunk_size
+            x = x.view(B * num_chunks, self.attn_chunk_size, C)
             new_B, new_T, C = x.size()
             q, k, v = self.receptance(x), self.key(x), self.value(x)
             k = k.view(new_B, new_T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)        
             q = q.view(new_B, new_T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
             v = v.view(new_B, new_T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
-            y = F.scaled_dot_product_attention(q, k, v, is_causal=True) # flash attention
+            y = F.scaled_dot_product_attention(q, k, v, is_causal=True) # here we use causal attention in chunk
             y = y.transpose(1, 2).contiguous().view(new_B, new_T, C) # re-assemble all head outputs side by side
             k = k.transpose(1, 2).contiguous().view(new_B, new_T, C)
             v = v.transpose(1, 2).contiguous().view(new_B, new_T, C)
             # output projection
             y = self.output(y)
             # unfold y, k, v into original shape
-            y = y.view(B, num_chunks*self.moba_chunk_size, C)
-            k = k.view(B, num_chunks*self.moba_chunk_size, C)
-            v = v.view(B, num_chunks*self.moba_chunk_size, C)
+            y = y.view(B, num_chunks*self.attn_chunk_size, C)
+            k = k.view(B, num_chunks*self.attn_chunk_size, C)
+            v = v.view(B, num_chunks*self.attn_chunk_size, C)
             # remove padding
-            if pad_size != self.moba_chunk_size:
+            if pad_size != self.attn_chunk_size:
                 y = y[:, :-pad_size, :]
                 k = k[:, :-pad_size, :]
                 v = v[:, :-pad_size, :]
@@ -433,14 +495,14 @@ class RWKV_CMix_x070(nn.Module):
         self.value = nn.Linear(n_embd * 4, n_embd, bias=False)
         self.x_k = nn.Parameter(torch.ones(1, 1, n_embd))
     
-    def forward(self, x, state):
+    def forward(self, x, x_prev):
         if len(x.shape) == 1:
-            xx = state["ffn_x_prev"] - x
+            xx = x_prev - x
             k = x + xx * self.x_k.squeeze()
             k = torch.relu(self.key(k)) ** 2
             return self.value(k), x
         
-        xx = torch.cat((state["ffn_x_prev"].unsqueeze(0), x[:-1,:])) - x
+        xx = torch.cat((x_prev.unsqueeze(0), x[:-1,:])) - x
         k = x + xx * self.x_k.squeeze(0) # (1, 1, C) -> (1, C)
         k = torch.relu(self.key(k)) ** 2
         return self.value(k), x[-1,:]
@@ -459,7 +521,7 @@ class SparseAttentionBlock(nn.Module):
             self.ln1(x), state['k_cache'], state['v_cache']
             )
         x = x + xx
-        xx, state['ffn_x_prev'] = self.ffn(self.ln2(x), state)
+        xx, state['ffn_x_prev'] = self.ffn(self.ln2(x), state['ffn_x_prev'])
         x = x + xx
         return x, state
     
@@ -471,32 +533,38 @@ class SparseAttentionBlock(nn.Module):
             self.ln1(x), state['k_cache'], state['v_cache']
             )
         x = x + xx
-        xx, state['ffn_x_prev'] = self.ffn(self.ln2(x), state)
+        xx, state['ffn_x_prev'] = self.ffn(self.ln2(x), state['ffn_x_prev'])
         x = x + xx
         return x, state
 
 @dataclass
 class RWKV_X_Config:
-    n_rwkv_layer: int
-    n_moba_layer: int
-    n_head: int
-    n_embd: int
-    moba_chunk_size: int = 2048
-    moba_topk: int = 3
+    n_rwkv_layer: int = 0
+    n_attn_layer: int = 0
+    n_head: int = 0
+    n_embd: int = 0
     head_size: int = 64
+    attn_chunk_size: int = 2000
+    attn_topk: int = 3
+    short_sequence_criteria: int = 8000 # if sequence length <= this, use full attention
+    max_kv_cache_size: int = 20000
+    kv_cache_window_size: int = 2000
+    min_kv_cache_size: int = 16000
+    prefill_attn_mode: str = 'chunk' # 'chunk' or 'full'
+    deocoding_attn_mode: str = 'full' # 'full' or 'snapKV' or 'new'
 
 class RWKV_X(nn.Module):
-    def __init__(self, model_path, strategy):
+    def __init__(self, model_path, strategy, config=None):
         super().__init__()
         print(f'Loading {model_path} ({strategy})\n')
-        rwkv_state_dict, moba_state_dict, config = self.load_from_ckpt(model_path, strategy)
+        rwkv_state_dict, attn_state_dict, config = self.load_from_ckpt(model_path, strategy, config)
         self.rwkv = RWKV_x070(rwkv_state_dict).to(device=DEVICE).to(DTYPE)
-        self.moba = nn.ModuleList([SparseAttentionBlock(config) for i in range(config.n_moba_layer)])
-        self.moba.to(device=DEVICE).to(DTYPE)
-        self.moba.load_state_dict(moba_state_dict, strict=True)
+        self.attn = nn.ModuleList([SparseAttentionBlock(config) for _ in range(config.n_attn_layer)])
+        self.attn.to(device=DEVICE).to(DTYPE)
+        self.attn.load_state_dict(attn_state_dict, strict=True)
         self.config = config
 
-    def load_from_ckpt(self, model_path, strategy):
+    def load_from_ckpt(self, model_path, strategy, config=None):
         global DTYPE, DEVICE
         ss = strategy.split(' ')
         DEVICE = ss[0]
@@ -516,21 +584,36 @@ class RWKV_X(nn.Module):
             raise FileNotFoundError(f"Model path {model_path} does not exist.")
 
         rwkv_state_dict = {k[5:]: v for k, v in ckpt.items() if k.startswith("rwkv.")}
-        moba_state_dict = {k[5:]: v for k, v in ckpt.items() if k.startswith("moba.")}
+        attn_state_dict = {k[5:]: v for k, v in ckpt.items() if k.startswith("moba.")}
 
         n_embd = rwkv_state_dict['emb.weight'].shape[1]
         n_head = n_embd // 64
         n_rwkv_layer = len({k.split('.')[1] for k in rwkv_state_dict if k.startswith("blocks.")})
-        n_moba_layer = len({k.split('.')[0] for k in moba_state_dict if k[0].isdigit()})
-        config = RWKV_X_Config(
-            n_rwkv_layer=n_rwkv_layer,
-            n_moba_layer=n_moba_layer,
-            n_head=n_head,
-            n_embd=n_embd,
-            moba_chunk_size=2048,
-            moba_topk=3
-        )
-        return rwkv_state_dict, moba_state_dict, config
+        n_attn_layer = len({k.split('.')[0] for k in attn_state_dict if k[0].isdigit()})
+        if config is None:
+            config = RWKV_X_Config(
+                n_rwkv_layer=n_rwkv_layer,
+                n_attn_layer=n_attn_layer,
+                n_head=n_head,
+                n_embd=n_embd,
+            )
+        else:
+            config = RWKV_X_Config(
+                n_rwkv_layer=n_rwkv_layer,
+                n_attn_layer=n_attn_layer,
+                n_head=n_head,
+                n_embd=n_embd,
+                attn_chunk_size=config.attn_chunk_size,
+                attn_topk=config.attn_topk,
+                max_kv_cache_size=config.max_kv_cache_size,
+                prefill_attn_mode=config.prefill_attn_mode,
+                deocoding_attn_mode=config.deocoding_attn_mode,
+                short_sequence_criteria=config.short_sequence_criteria,
+                kv_cache_window_size=config.kv_cache_window_size,
+                min_kv_cache_size=config.min_kv_cache_size,
+                head_size=config.head_size
+            )
+        return rwkv_state_dict, attn_state_dict, config
 
 
     @torch.inference_mode()
@@ -541,14 +624,14 @@ class RWKV_X(nn.Module):
                 rwkv_state[i*3+0] = torch.zeros(self.config.n_embd, dtype=DTYPE, requires_grad=False, device=DEVICE)
                 rwkv_state[i*3+1] = torch.zeros((self.config.n_embd // self.config.head_size, self.config.head_size, self.config.head_size), dtype=torch.float, requires_grad=False, device=DEVICE)
                 rwkv_state[i*3+2] = torch.zeros(self.config.n_embd, dtype=DTYPE, requires_grad=False, device=DEVICE)
-            moba_state = []
-            for _ in range(self.config.n_moba_layer):
-                moba_state.append({
+            attn_state = []
+            for _ in range(self.config.n_attn_layer):
+                attn_state.append({
                     'k_cache': torch.zeros((1, 0, self.config.n_embd), dtype=DTYPE, device=DEVICE, requires_grad=False),
                     'v_cache': torch.zeros((1, 0, self.config.n_embd), dtype=DTYPE, device=DEVICE, requires_grad=False),
                     'ffn_x_prev':  torch.zeros(self.config.n_embd, dtype=DTYPE, requires_grad=False, device=DEVICE)
                 })
-            state = {'rwkv': rwkv_state, 'moba': moba_state}
+            state = {'rwkv': rwkv_state, 'attn': attn_state}
 
         if type(idx) is list:
             if len(idx) > 1:
@@ -563,14 +646,14 @@ class RWKV_X(nn.Module):
         x = z['emb.weight'][idx]
         v_first = torch.empty_like(x)
 
-        rwkv_id, moba_id = 0, 0
+        rwkv_id, attn_id = 0, 0
         for block in self.get_block_exe_order():
             if isinstance(block, RWKVBlock):
                 x, state['rwkv'], v_first = block.forward_one(x, state['rwkv'], v_first)
                 rwkv_id += 1
             else:
-                x, state['moba'][moba_id] = block.forward_one(x, state['moba'][moba_id])
-                moba_id += 1
+                x, state['attn'][attn_id] = block.forward_one(x, state['attn'][attn_id])
+                attn_id += 1
 
         x = F.layer_norm(x, (self.config.n_embd,), weight=z['ln_out.weight'], bias=z['ln_out.bias'])
         x = x @ z['head.weight']
@@ -579,17 +662,16 @@ class RWKV_X(nn.Module):
     def forward_seq(self, idx: List[int], state: dict, full_output=False):
         z = self.rwkv.z
         x = z['emb.weight'][idx]
-        x_prev = x[0].clone()
         v_first = torch.empty_like(x)
 
-        rwkv_id, moba_id = 0, 0
+        rwkv_id, attn_id = 0, 0
         for block in self.get_block_exe_order():
             if isinstance(block, RWKVBlock):
                 x, state['rwkv'], v_first = block.forward_seq(x, state['rwkv'], v_first)
                 rwkv_id += 1
             else:
-                x, state['moba'][moba_id] = block.forward_seq(x, state['moba'][moba_id])
-                moba_id += 1
+                x, state['attn'][attn_id] = block.forward_seq(x, state['attn'][attn_id])
+                attn_id += 1
 
         if not full_output:
             x = x[-1]
@@ -598,14 +680,14 @@ class RWKV_X(nn.Module):
         return x, state
     
     def get_block_exe_order(self):
-        if self.config.n_moba_layer == 0:
+        if self.config.n_attn_layer == 0:
             return self.rwkv.blocks
-        if self.config.n_moba_layer == 1:
-            blocks = self.rwkv.blocks + self.moba
+        if self.config.n_attn_layer == 1:
+            blocks = self.rwkv.blocks + self.attn
             return blocks
-        interval = len(self.rwkv.blocks) // self.config.n_moba_layer # 12 // 4 = 3
-        blocks = [] # [RWKVBlock * interval, MOBABlock, RWKVBlock * interval, MOBABlock, ...]
-        for i in range(self.config.n_moba_layer):
+        interval = len(self.rwkv.blocks) // self.config.n_attn_layer # 12 // 4 = 3
+        blocks = [] # [RWKVBlock * interval, AttnBlock, RWKVBlock * interval, AttnBlock, ...]
+        for i in range(self.config.n_attn_layer):
             blocks += self.rwkv.blocks[i * interval: (i + 1) * interval]
-            blocks.append(self.moba[i])
+            blocks.append(self.attn[i])
         return blocks
